<?php
/**
 * Copyright © 2015 Magenest. All rights reserved.
 * See COPYING.txt for license details.
 *
 * Magenest_Salesforce extension
 * NOTICE OF LICENSE
 *
 * @category Magenest
 * @package  Magenest_Salesforce
 * @author   ThaoPV
 */

namespace Magenest\Salesforce\Model\Sync;

use Magenest\Salesforce\Helper\Data as HelperData;
use Magenest\Salesforce\Logger\Logger;
use Magenest\Salesforce\Model\Connector;
use Magenest\Salesforce\Model\Data;
use Magenest\Salesforce\Model\ReportFactory as ReportFactory;
use Magenest\Salesforce\Model\RequestLogFactory;
use Magenest\Salesforce\Model\ResourceModel\Queue as QueueResource;
use Magenest\Salesforce\Model\ResourceModel\Queue\CollectionFactory as QueueCollectionFactory;
use Magenest\Salesforce\Model\ResourceModel\Report\CollectionFactory as ReportCollectionFactory;
use Magento\Config\Model\ResourceModel\Config as ResourceModelConfig;
use Magento\Customer\Model\CustomerRegistry;
use Magento\Customer\Model\ResourceModel\Customer;
use Magento\Framework\App\Config\ScopeConfigInterface as ScopeConfigInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\HTTP\ClientFactory;
use Magento\Framework\Model\ResourceModel\Iterator;
use Magento\Framework\Serialize\Serializer\Json;
use Magento\Framework\Serialize\Serializer\Serialize;

class Account extends Connector
{
    const SALESFORCE_ACCOUNT_ATTRIBUTE_CODE = 'salesforce_account_id';

    /** @var Customer */
    protected $customerResource;

    /** @var CustomerRegistry */
    private $customerRegistry;

    /** @var Job */
    protected $_job;

    /** @var Data */
    protected $_data;

    /** @var DataGetter */
    protected $dataGetter;

    /** @var HelperData */
    protected $_helper;

    //Personal Account Type Id
    protected $recordedTypeId = null;

    protected $existedAccounts = null;

    protected $createAccountIds = null;

    protected $updateAccountIds = null;

    /**
     * Account constructor.
     * @param Data $data
     * @param Job $job
     * @param DataGetter $dataGetter
     * @param ReportCollectionFactory $reportCollectionFactory
     * @param HelperData $helperData
     * @param Json $json
     * @param Serialize $serialize
     * @param ScopeConfigInterface $scopeConfig
     * @param ResourceModelConfig $resourceConfig
     * @param ReportFactory $reportFactory
     * @param QueueCollectionFactory $queueCollectionFactory
     * @param QueueResource $queueResource
     * @param RequestLogFactory $requestLogFactory
     * @param Customer $customerResource
     * @param CustomerRegistry $customerRegistry
     * @param ClientFactory $client
     * @param Iterator $iterator
     * @param Logger $logger
     */
    public function __construct(
        Data                    $data,
        Job                     $job,
        DataGetter              $dataGetter,
        ReportCollectionFactory $reportCollectionFactory,
        HelperData              $helperData,
        Json                    $json,
        Serialize               $serialize,
        ScopeConfigInterface    $scopeConfig,
        ResourceModelConfig     $resourceConfig,
        ReportFactory           $reportFactory,
        QueueCollectionFactory  $queueCollectionFactory,
        QueueResource           $queueResource,
        RequestLogFactory       $requestLogFactory,
        Customer                $customerResource,
        CustomerRegistry        $customerRegistry,
        ClientFactory           $client,
        Iterator                $iterator,
        Logger                  $logger
    ) {
        $this->_data            = $data;
        $this->_type            = 'Account';
        $this->_table           = 'customer';
        $this->_job             = $job;
        $this->dataGetter       = $dataGetter;
        $this->customerResource = $customerResource;
        $this->customerRegistry = $customerRegistry;
        parent::__construct(
            $reportCollectionFactory,
            $helperData,
            $json,
            $serialize,
            $scopeConfig,
            $resourceConfig,
            $reportFactory,
            $queueCollectionFactory,
            $queueResource,
            $requestLogFactory,
            $iterator,
            $client,
            $logger
        );
    }

    /**
     * Update or create new a record
     *
     * @param int $id
     * @return string
     * @throws LocalizedException
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     */
    public function sync($id, $isSubscriber = false)
    {
        $customer = $this->customerRegistry->retrieve($id);
        $email    = $customer->getEmail();
        $id       = $this->massCheckRecord([$id]);
        $id       = is_array($id) ? reset($id) : $id;

        // Pass data of customer to array
        $params = $this->_data->getCustomer($customer, $this->_type);
        if ($this->_helperData->enablePersonalAccount()) {
            if ($this->recordedTypeId == null) {
                $this->recordedTypeId = $this->getRecordTypeId();
            }
            if (isset($params['Name'])) {
                unset($params['Name']);
            }
            $params = array_replace($params, [
                'RecordTypeId' => $this->recordedTypeId,
                'FirstName'    => $customer->getFirstname()
                    ?? $this->_data->getDefaultAccountName('FirstName')
                    ?? $email,
                'LastName'     => $customer->getLastname()
                    ?? $this->_data->getDefaultAccountName('LastName')
                    ?? $email,
                'PersonEmail'  => $email
            ]);
        } else {
            $params = array_replace($params, [
                'Name' => $email
            ]);
        }
        if (is_array($isSubscriber) && isset($isSubscriber['field']) && isset($isSubscriber['status'])) {
            $params = array_replace($params, [
                $isSubscriber['field'] => $isSubscriber['status']
            ]);
        }
        $this->_helperData->saveLog("Account params mapping: " . $this->_json->serialize($params));
        if (!empty($id)) {
            $this->updateRecords($this->_type, $id, $params, $customer->getId());
        } else {
            $id = $this->createRecords($this->_type, $params, $customer->getId());
        }
        $this->saveAttribute($customer, $id);

        return $id;
    }

    /**
     * @param $email
     * @param $firstName
     * @param $lastName
     * @return bool|mixed
     * @throws \Zend_Http_Client_Exception
     * @throws \Exception
     */
    public function syncByEmail($email, $firstName = null, $lastName = null)
    {
        $params = [];
        $id     = $this->searchRecords($this->_type, $this->_helperData->enablePersonalAccount() ? 'PersonEmail' : 'Name', $email);
        if (!$id) {
            if ($this->_helperData->enablePersonalAccount()) {
                if ($this->recordedTypeId == null) {
                    $this->recordedTypeId = $this->getRecordTypeId();
                }
                $params = array_replace($params, [
                    'RecordTypeId' => $this->recordedTypeId,
                    'FirstName'    => $firstName
                        ?? $this->_data->getDefaultAccountName('FirstName')
                        ?? $email,
                    'LastName'     => $lastName
                        ?? $this->_data->getDefaultAccountName('LastName')
                        ?? $email,
                    'PersonEmail'  => $email
                ]);
            } else {
                $params = array_replace($params, [
                    'Name' => $email
                ]);
            }
            $id = $this->createRecords($this->_type, $params);
        }

        return $id;
    }

    /**
     * @return array|mixed|string|null
     */
    public function syncQueue()
    {
        $createResponse = $this->createAccounts();
        $updateResponse = $this->updateAccounts();
        $response       = $createResponse + $updateResponse;
        $this->unsetCreateQueue();
        $this->unsetUpdateQueue();
        return $response;
    }

    /**
     * Send request to create accounts
     */
    protected function createAccounts()
    {
        $response = [];
        if (!empty($this->createAccountIds)) {
            $response = $this->sendAccountsRequest($this->createAccountIds, 'insert');
        }
        return $response;
    }

    /**
     * Send request to update accounts
     */
    protected function updateAccounts()
    {
        $response = [];
        if (!empty($this->updateAccountIds)) {
            $response = $this->sendAccountsRequest($this->updateAccountIds, 'update');
        }
        return $response;
    }

    /**
     * @param $customerId
     * @throws LocalizedException
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     * @throws \Zend_Http_Client_Exception
     */
    public function addRecord($customerId)
    {
        $customer = $this->customerRegistry->retrieve($customerId);
        $id       = $this->checkSalesforceCustomerStatus($customer);
        if (!$id) {
            $this->addToCreateQueue($customer);
        } else {
            $this->addToUpdateQueue($customer, $id);
        }
    }

    /**
     * @param $customerIds
     * @return array
     * @throws LocalizedException
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     */
    public function massCheckRecord($customerIds)
    {
        $queueCount      = $recordCount = 0;
        $response        = [];
        $queryLimit      = 12000;
        $recordLimit     = 250;
        $isPersonAccount = $this->_helperData->enablePersonalAccount();
        $count           = count($customerIds);
        if ($isPersonAccount) {
            $query = "SELECT Id, PersonEmail FROM Account WHERE ";
            foreach ($customerIds as $customerId) {
                $queueCount++;
                $recordCount++;
                $customer = $this->customerRegistry->retrieve($customerId);

                //$query .= "(FirstName = '" . addslashes($customer->getFirstname()) . "' AND ";
                //$query .= "LastName = '" . addslashes($customer->getLastname()) . "' AND ";
                $query .= "(PersonEmail = '" . addslashes($customer->getEmail()) . "') ";

                if (strlen($query) >= $queryLimit || $queueCount >= $recordLimit) {
                    $query .= ' LIMIT ' . $recordLimit;
                    $path  = '/services/data/v' . self::SALES_FORCE_API_VERSION . '/query?q=' . urlencode($query);

                    $response += $this->sendRequestUsingCurl($path, 'PersonEmail');

                    $queueCount = 0;
                    $query      = "SELECT Id, PersonEmail FROM Account WHERE ";
                } elseif ($recordCount !== $count) {
                    $query .= "OR ";
                }
            }
        } else {
            $query = "SELECT Id, Name FROM Account WHERE ";
            foreach ($customerIds as $customerId) {
                $queueCount++;
                $recordCount++;
                $customer = $this->customerRegistry->retrieve($customerId);

                $query .= "(Name = '" . $customer->getEmail() . "') ";

                if (strlen($query) >= $queryLimit || $queueCount >= $recordLimit) {
                    $query .= ' LIMIT ' . $recordLimit;
                    $path  = '/services/data/v' . self::SALES_FORCE_API_VERSION . '/query?q=' . urlencode($query);

                    $response += $this->sendRequestUsingCurl($path);

                    $queueCount = 0;
                    $query      = "SELECT Id, Name FROM Account WHERE ";
                } elseif ($recordCount !== $count) {
                    $query .= "OR ";
                }
            }
        }

        if ($query !== "SELECT Id, PersonEmail FROM Account WHERE " || $query !== "SELECT Id, Name FROM Account WHERE ") {
            $query .= ' LIMIT ' . $recordLimit;
            $path  = '/services/data/v' . self::SALES_FORCE_API_VERSION . '/query?q=' . urlencode($query);

            $response += $this->sendRequestUsingCurl($path, $isPersonAccount ? "PersonEmail" : "Name");
        }

        return $response;
    }

    /**
     * @param $customerId
     * @return array|false|mixed|void
     */
    private function getStoredSalesforceId($customerId)
    {
        $allReport = $this->getLogReport();
        if (isset($allReport[$customerId]['record_id']) && isset($allReport[$customerId]['datetime'])) {
            $diff = date_create($allReport[$customerId]['datetime'])->diff(date_create());
            return (($diff->d * 24) + $diff->h + ($diff->i / 60)) > 1 ? $allReport[$customerId]['record_id'] : false;
        }
    }

    /**
     * @param \Magento\Customer\Model\Customer $customer
     */
    protected function addToCreateQueue($customer)
    {
        $this->createAccountIds[] = [
            'mObj' => $customer,
            'mid'  => $customer->getId()
        ];
    }

    /**
     * @param \Magento\Customer\Model\Customer $customer
     * @param $salesforceId
     */
    protected function addToUpdateQueue($customer, $salesforceId)
    {
        $this->updateAccountIds[] = [
            'mObj' => $customer,
            'mid'  => $customer->getId(),
            'sid'  => $salesforceId
        ];
    }

    protected function unsetCreateQueue()
    {
        $this->createAccountIds = null;
    }

    protected function unsetUpdateQueue()
    {
        $this->updateAccountIds = null;
    }

    /**
     * sync array of emails
     *
     * @param array $emails
     * @param string $operation
     * @return array
     * @throws \Exception
     */
    public function syncByEmails($emails, $operation)
    {
        if (empty($emails)) {
            return [];
        }
        $ids            = [];
        $params         = [];
        $unsyncedEmails = [];
        foreach ($emails as $email) {
            if (isset($email['email'])) {
                $sfId = $this->getCustomerByEmail($email['email']);
                if ($sfId) {
                    foreach ($email['order_ids'] as $orderId) {
                        $ids[$orderId] = $sfId;
                    }
                } else {
                    if ($this->_helperData->enablePersonalAccount()) {
                        if ($this->recordedTypeId == null) {
                            $this->recordedTypeId = $this->getRecordTypeId();
                        }
                        $params = array_replace($params, [
                            'RecordTypeId' => $this->recordedTypeId,
                            'FirstName'    => $email['firstName']
                                ?? $this->_data->getDefaultAccountName('FirstName')
                                ?? $email['email'],
                            'LastName'     => $email['lastName']
                                ?? $this->_data->getDefaultAccountName('LastName')
                                ?? $email['email'],
                            'PersonEmail'  => $email['email']
                        ]);
                    } else {
                        $params = [
                            'Name' => $email['email']
                        ];
                    }
                    $unsyncedEmails[] = $params;
                }
            }
        }

        if (!empty($unsyncedEmails)) {
            $response = $this->_job->sendBatchRequest(
                $operation,
                $this->_type,
                $this->_json->serialize($unsyncedEmails)
            );
            $this->saveReports($operation, $this->_type, $response, $unsyncedEmails, $unsyncedEmails);
            if (is_array($response)) {
                $total = count($response);
                for ($i = 0; $i < $total; $i++) {
                    if (isset($response[$i]['success']) && $response[$i]['success']) {
                        foreach ($emails[$unsyncedEmails[$i]['Name']]['order_ids'] as $order_id) {
                            $ids[$order_id] = $response[$i]['id'];
                        }
                    }
                }
            }
        }
        return $ids;
    }

    /**
     * @param array $accountIds
     * @param string $operation
     * @return mixed|string
     * @throws \Exception
     */
    public function sendAccountsRequest($accountIds, $operation)
    {
        $params = [];
        foreach ($accountIds as $accountId) {
            $customer = $accountId['mObj'];
            $info     = $this->_data->getCustomer($customer, $this->_type);
            if (isset($accountId['sid'])) {
                $info += ['Id' => $accountId['sid']];
            }
            if ($this->_helperData->enablePersonalAccount()) {
                if (isset($info['Name'])) {
                    unset($info['Name']);
                }
                if ($this->recordedTypeId == null) {
                    $this->recordedTypeId = $this->getRecordTypeId();
                }
                $info = array_replace($info, [
                    'RecordTypeId' => $this->recordedTypeId,
                    'FirstName'    => $customer->getFirstname()
                        ?? $this->_data->getDefaultAccountName('FirstName')
                        ?? $customer->getEmail(),
                    'LastName'     => $customer->getLastname()
                        ?? $this->_data->getDefaultAccountName('LastName')
                        ?? $customer->getEmail(),
                    'PersonEmail'  => $customer->getEmail()
                ]);
            } else {
                $info = array_replace($info, [
                    'Name' => $customer->getEmail()
                ]);
            }
            $params[] = $info;
        }
        $response = $this->_job->sendBatchRequest($operation, $this->_type, $this->_json->serialize($params));
        $this->saveReports($operation, $this->_type, $response, $accountIds, $params);
        return $response;
    }

    /**
     * @return array|mixed|string|null
     * @throws \Exception
     */
    public function massSyncQueue()
    {
        $response = [];
        $type     = $this->_type;
        $queue    = $this->_queueResource->fetchEntityIdByType($type);
        if (!empty($queue)) {
            $update = $this->massCheckRecord($queue);

            $count              = 0;
            $maxRecord          = 50;
            $queueCollectionArr = [];
            foreach ($queue as $queueId => $customerId) {
                $customer = $this->customerRegistry->retrieve($customerId);
                if (isset($update[$customer->getEmail()]) && $update[$customer->getEmail()] === false) {
                    $queueCollectionArr[] = $queueId;
                    continue;
                } elseif (!empty($update[$customer->getEmail()])) {
                    $this->addToUpdateQueue($customer, $update[$customer->getEmail()]);
                } else {
                    $this->addToCreateQueue($customer);
                }

                $count++;
                $queueCollectionArr[] = $queueId;
                if ($count >= $maxRecord) {
                    $response += $this->syncQueue();
                    $count    = 0;

                    $this->_queueResource->deleteMultiQueues($queueCollectionArr);
                    $queueCollectionArr = [];
                }
            }

            $response += $this->syncQueue();
            $this->_queueResource->deleteMultiQueues($queueCollectionArr);
        }

        return $response;
    }

    /**
     * @param $customerIds
     * @param $response
     * @throws \Exception
     */
    protected function saveAttributes($customerIds, $response)
    {
        if (empty($customerIds)) {
            return;
        }
        if (is_array($response) && is_array($customerIds)) {
            $total = count($response);
            for ($i = 0; $i < $total; $i++) {
                if (isset($customerIds[$i])) {
                    $customer = $customerIds[$i]['mObj'];
                    if (isset($response[$i]['id']) && $customer->getId()) {
                        $this->saveAttribute($customer, $response[$i]['id']);
                    }
                }
            }
        }
    }

    /**
     * @param $customer
     * @param $salesforceId
     * @return void
     */
    protected function saveAttribute($customer, $salesforceId)
    {
        try {
            if (is_array($salesforceId)) {
                $salesforceId = reset($salesforceId);
            }
            $customerData = $customer->getDataModel();
            $customerData->setId($customer->getId());
            $customerData->setCustomAttribute(self::SALESFORCE_ACCOUNT_ATTRIBUTE_CODE, $salesforceId);
            $customer->updateData($customerData);
            $this->customerResource->saveAttribute($customer, self::SALESFORCE_ACCOUNT_ATTRIBUTE_CODE);
        } catch (\Throwable $e) {
            $this->_logger->critical($e->getMessage(), ['trace' => $e->getTraceAsString()]);
        }
    }
}
