<?php
namespace Magenest\Salesforce\Model\Sync;

use Magenest\Salesforce\Logger\Logger;
use Magenest\Salesforce\Model\Connector;
use Magenest\Salesforce\Model\Data;
use Magenest\Salesforce\Model\ReportFactory as ReportFactory;
use Magenest\Salesforce\Model\RequestLogFactory;
use Magenest\Salesforce\Model\ResourceModel\Queue as QueueResource;
use Magenest\Salesforce\Model\ResourceModel\Queue\CollectionFactory as QueueCollectionFactory;
use Magenest\Salesforce\Model\ResourceModel\Report\CollectionFactory as ReportCollectionFactory;
use Magento\Config\Model\ResourceModel\Config as ResourceModelConfig;
use Magento\Customer\Model\CustomerFactory;
use Magento\Framework\App\Config\ScopeConfigInterface as ScopeConfigInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\HTTP\ClientFactory;
use Magento\Framework\Model\ResourceModel\Iterator;
use Magento\Framework\Serialize\Serializer\Serialize;
use Magento\Sales\Model\Order\Item;
use Magento\Sales\Model\OrderFactory;

class Order extends Connector
{
    const SALESFORCE_ORDER_ATTRIBUTE_CODE = 'salesforce_order_id';

    /** @var OrderFactory */
    protected $_orderFactory;

    /** @var CustomerFactory */
    protected $_customerFactory;

    /** @var Account */
    protected $_account;

    /** @var Contact */
    protected $_contact;

    /** @var Product */
    protected $_product;

    /** @var Job */
    protected $_job;

    /** @var Data */
    protected $_data;

    /** @var \Magento\Catalog\Model\CategoryFactory */
    protected $_category;

    /** @var DataGetter */
    protected $dataGetter;

    /** @var array|null */
    protected $createOrderIds = null;

    /** @var array|null */
    protected $updateOrderIds = null;

    /**
     * Order constructor.
     * @param OrderFactory $orderFactory
     * @param CustomerFactory $customer
     * @param Account $account
     * @param Contact $contact
     * @param Product $product
     * @param Data $data
     * @param Job $job
     * @param DataGetter $dataGetter
     * @param \Magenest\Salesforce\Helper\Data $helperData
     * @param \Magento\Catalog\Model\CategoryFactory $category
     * @param ReportCollectionFactory $reportCollectionFactory
     * @param \Magento\Framework\Serialize\Serializer\Json $json
     * @param Serialize $serialize
     * @param ScopeConfigInterface $scopeConfig
     * @param ResourceModelConfig $resourceConfig
     * @param ReportFactory $reportFactory
     * @param QueueCollectionFactory $queueCollectionFactory
     * @param QueueResource $queueResource
     * @param RequestLogFactory $requestLogFactory
     * @param ClientFactory $client
     * @param Iterator $iterator
     * @param Logger $logger
     */
    public function __construct(
        OrderFactory                                      $orderFactory,
        CustomerFactory                                   $customer,
        Account                                           $account,
        Contact                                           $contact,
        Product                                           $product,
        Data                                              $data,
        Job                                               $job,
        DataGetter                                        $dataGetter,
        \Magenest\Salesforce\Helper\Data                  $helperData,
        \Magento\Catalog\Model\CategoryFactory            $category,
        ReportCollectionFactory                           $reportCollectionFactory,
        \Magento\Framework\Serialize\Serializer\Json      $json,
        \Magento\Framework\Serialize\Serializer\Serialize $serialize,
        ScopeConfigInterface                              $scopeConfig,
        ResourceModelConfig                               $resourceConfig,
        ReportFactory                                     $reportFactory,
        QueueCollectionFactory                            $queueCollectionFactory,
        QueueResource                                     $queueResource,
        RequestLogFactory                                 $requestLogFactory,
        ClientFactory                                     $client,
        Iterator                                          $iterator,
        Logger                                            $logger
    ) {
        $this->_orderFactory    = $orderFactory;
        $this->_customerFactory = $customer;
        $this->_account         = $account;
        $this->_contact         = $contact;
        $this->_product         = $product;
        $this->_data            = $data;
        $this->_type            = 'Order';
        $this->_table           = 'order';
        $this->_job             = $job;
        $this->dataGetter       = $dataGetter;
        $this->_category        = $category;
        parent::__construct(
            $reportCollectionFactory,
            $helperData,
            $json,
            $serialize,
            $scopeConfig,
            $resourceConfig,
            $reportFactory,
            $queueCollectionFactory,
            $queueResource,
            $requestLogFactory,
            $iterator,
            $client,
            $logger
        );
    }

    /**
     * @param $d
     *
     * @return array|bool|string
     */
    public function utf8ize($d)
    {
        if (is_array($d)) {
            foreach ($d as $k => $v) {
                $d[$k] = $this->utf8ize($v);
            }
        } elseif (is_string($d)) {
            return utf8_encode($d);
        }
        return $d;
    }

    /**
     * @param $incrementId
     * @return bool|mixed|string
     * @throws LocalizedException
     * @throws \Zend_Http_Client_Exception
     * @throws \Exception
     */
    public function sync($incrementId)
    {
        $order      = $this->_orderFactory->create()->loadByIncrementId($incrementId);
        $customerId = $order->getCustomerId();
        $date       = date('Y-m-d', strtotime($order->getCreatedAt()));
        $email      = $order->getCustomerEmail();
        $params     = $this->_data->getOrder($order, $this->_type);
        // Get pricebookId of "Standard Price Book"
        $pricebookId = $this->_getStandardPriceBookId();

        /**
         * Require Field:
         *
         * 1. AccountId
         * 2. EffectiveDate
         * 3. Status
         * 4. PriceBook2Id
         */
        $params       += [
            'EffectiveDate' => $date,
            'Status'        => 'Draft'
        ];
        $params       = array_replace($params, ['Pricebook2Id' => $pricebookId]);
        $salesforceId = $this->checkExistedOrder($order);
        if ($salesforceId) {
            $paramsUTF8 = $this->utf8ize($params);
            $orderId    = $this->updateRecords($this->_type, $salesforceId, $paramsUTF8, $order->getIncrementId());
            $this->saveAttribute($order, $orderId);
            $syncSide = false;
        } else {
            /*
             * 1. Get accountId, create new if not exist
             * 2. Create new Contacts if not exist
            */
            if ($customerId) {
                $accountId = $this->_account->sync($customerId);
                $this->_contact->sync($customerId);
            } else {
                $accountId = $this->_account->syncByEmail(
                    $email,
                    $order->getCustomerFirstname(),
                    $order->getCustomerLastname()
                );
                $data      = [
                    'Email' => $email
                ];

                /**
                 * Required field
                 * If customer hasn't created an Account, LastName will be Guest
                 */
                $data += [
                    'LastName' => 'Guest'
                ];
                $this->_contact->syncByEmail($data);
            }
            $params     = array_replace($params, ['AccountId' => $accountId]);
            $paramsUTF8 = $this->utf8ize($params);
            $this->_helperData->saveLog("Order params mapping: " . $this->_json->serialize($params));
            // Create new Order
            $orderId = $this->createRecords($this->_type, $paramsUTF8, $order->getIncrementId());
            $this->saveAttribute($order, $orderId);
            $syncSide = true;
        }

        /**
         * Add new record to OrderItem need:
         *
         * 1. ProductId
         * 2. PricebookEntryId
         */
        $this->syncOrderItems($order, $pricebookId, $orderId, $syncSide);
        return $orderId;
    }

    /**
     * @param \Magento\Sales\Model\Order $order
     * @param $pricebook2Id
     * @param $orderId
     * @param $syncSide
     * @throws LocalizedException
     * @throws \Zend_Http_Client_Exception
     */
    private function syncOrderItems(\Magento\Sales\Model\Order $order, $pricebook2Id, $orderId, $syncSide)
    {
        if (!empty($orderId)) {
            /** @var Item $item */
            foreach ($order->getAllVisibleItems() as $item) {
                $product         = $item->getProduct();
                $productEntityId = $item->getProductId();
                $qty             = $item->getQtyOrdered();
                $price           = $item->getPrice();
                if ($price > 0) {
                    $productId = $this->_product->syncOrderItem($item);
                    $pricebookEntryId = $this->getPricebookEntryId(
                        $productId,
                        $pricebook2Id,
                        $product ? ($product->getPrice() ?: 0) : ($item->getOriginalPrice() ?: 0),
                        $item->getProductId()
                    );

                    if ($productId && $pricebookEntryId) {
                        if (!$this->searchOrderItems('OrderItem', 'OrderId', $orderId, $pricebookEntryId, (double)$qty, (double)$price)) {
                            $output = [
                                'PricebookEntryId' => $pricebookEntryId,
                                'OrderId'          => $orderId,
                                'Quantity'         => $qty,
                                'UnitPrice'        => $price
                            ];
                            $this->_helperData->saveLog("OrderItem params mapping: " . $this->_json->serialize($output));
                            $this->createRecords('OrderItem', $output, $productEntityId);
                        }
                    }
                }
            }

            if ($syncSide) {
                if ($taxInfo = $this->getTaxItemInfo($order, $orderId)) {
                    $this->createRecords('OrderItem', $taxInfo, 'TAX');
                }

                if ($shippingInfo = $this->getShippingItemInfo($order, $orderId)) {
                    $this->createRecords('OrderItem', $shippingInfo, 'SHIPPING');
                }

                if ($discountInfo = $this->getDiscountItemInfo($order, $orderId)) {
                    $this->createRecords('OrderItem', $discountInfo, 'DISCOUNT');
                }
            }
        }
    }

    /**
     * @return array|mixed|string|null
     */
    public function syncAllOrders()
    {
        try {
            $orders      = $this->_orderFactory->create()->getCollection();
            $lastOrderId = $orders->getLastItem()->getId();
            $count       = 0;
            $response    = [];
            foreach ($orders as $order) {
                $this->addRecord($order->getIncrementId());
                $count++;
                if ($count >= 10000 || $order->getId() == $lastOrderId) {
                    $response += $this->syncQueue();
                }
            }
            return $response;
        } catch (\Throwable $e) {
            $this->_logger->debug($e->getMessage());
        }
        return null;
    }

    /**
     * @param $orderIncrementId
     * @throws \Exception
     */
    public function addRecord($orderIncrementId)
    {
        $order        = $this->_orderFactory->create()->loadByIncrementId($orderIncrementId);
        $salesforceId = $this->checkExistedOrder($order);
        if ($order->getIncrementId()) {
            if (!$salesforceId) {
                $this->addToCreateOrderQueue($order);
            } else {
                $this->addToUpdateOrderQueue($order, $salesforceId);
            }
        } else {
            $this->saveReport(null, 'insert', $this->_type, 2, __("The Order with this increment id failed to load."), $orderIncrementId);
        }
    }

    /**
     * @return array|mixed|string|null
     * @throws \Exception
     */
    public function syncQueue()
    {
        $response = $this->createOrders();
        $this->saveAttributes($this->createOrderIds, $response);
        $response += $this->updateOrders();
        $response += $this->createOrderItems();
        $this->unsetCreateProductQueue();
        $this->unsetUpdateOrderQueue();
        return $response;
    }

    /**
     * @param \Magento\Sales\Model\Order $order
     */
    protected function addToCreateOrderQueue($order)
    {
        $this->createOrderIds[] = [
            'mid'  => $order->getIncrementId(),
            'mObj' => $order
        ];
    }

    /**
     * @param $order
     * @param $sId
     */
    protected function addToUpdateOrderQueue($order, $sId)
    {
        $this->updateOrderIds[] = [
            'mid'  => $order->getIncrementId(),
            'sid'  => $sId,
            'mObj' => $order
        ];
    }

    protected function unsetCreateProductQueue()
    {
        $this->createOrderIds = null;
    }

    protected function unsetUpdateOrderQueue()
    {
        $this->updateOrderIds = null;
    }

    /**
     * @return array|mixed|string
     * @throws LocalizedException|\Zend_Http_Client_Exception
     */
    protected function createOrders()
    {
        if ($this->createOrderIds === null) {
            return [];
        }
        $params      = [];
        $pricebookId = $this->_getStandardPriceBookId();

        $accountIds = $this->getCustomerWithOrder();
        $orderIds   = $this->createOrderIds;
        foreach ($orderIds as $key => $orderId) {
            /** @var \Magento\Sales\Model\Order $order */
            $order = $orderId['mObj'];
            $date  = date('Y-m-d', strtotime($order->getCreatedAt()));
            $info  = $this->_data->getOrder($order, $this->_type);
            if (empty($accountIds[$order->getId()])) {
                $this->saveReport(
                    null,
                    'insert',
                    $this->_type,
                    2,
                    __("Can't find AccountId associated with this Order. Try syncing the Account first."),
                    $order->getIncrementId(),
                    [
                        'EffectiveDate' => $date,
                        'Status'        => 'Draft',
                        'Pricebook2Id'  => $pricebookId
                    ]
                );
                unset($orderIds[$key]);
                continue;
            }

            /*
                * Require Field:
                *
                * 1. AccountId
                * 2. EffectiveDate
                * 3. Status
                * 4. PriceBook2Id
             */
            $info       += [
                'EffectiveDate' => $date,
                'Status'        => 'Draft'
            ];
            $info       = array_replace($info, [
                'Pricebook2Id'  => $pricebookId,
                'AccountId'     => $accountIds[$order->getId()]
            ]);
            $params[] = $info;
        }
        $paramsUTF8 = $this->utf8ize($params);
        $response   = $this->_job->sendBatchRequest('insert', $this->_type, $this->_json->serialize($paramsUTF8));
        $this->saveReports('create', $this->_type, $response, array_values($orderIds), $paramsUTF8);
        return $response;
    }

    /**
     * @return array|mixed|string
     * @throws LocalizedException|\Zend_Http_Client_Exception
     */
    public function updateOrders()
    {
        if ($this->updateOrderIds === null) {
            return [];
        }
        $params      = [];
        $orderIds    = $this->updateOrderIds;
        $pricebookId = $this->_getStandardPriceBookId();
        foreach ($orderIds as $orderId) {
            /** @var \Magento\Sales\Model\Order $order */
            $order    = $orderId['mObj'];
            $date     = date('Y-m-d', strtotime($order->getCreatedAt()));
            $info     = $this->_data->getOrder($order, $this->_type);
            $info     += [
                'EffectiveDate' => $date,
                'Status'        => 'Draft'
            ];
            $info     = array_replace($info, [
                'Pricebook2Id'  => $pricebookId,
                'Id'            => $orderId['sid']
            ]);
            $params[] = $info;
        }
        $response = $this->_job->sendBatchRequest('update', $this->_type, $this->_json->serialize($params));
        $this->saveReports('update', $this->_type, $response, $this->updateOrderIds, $params);
        return $response;
    }

    /**
     * @return array
     * @throws \Exception
     */
    protected function getCustomerWithOrder()
    {
        $accountIds               = [];
        $unsyncedOrderCustomerIds = [];
        $unsyncedEmails           = [];
        $unsyncedDataEmails       = [];
        $orderIds                 = $this->createOrderIds;
        foreach ($orderIds as $orderId) {
            /** @var \Magento\Sales\Model\Order $order */
            $order      = $orderId['mObj'];
            $orderId    = $order->getId();
            $customerId = $order->getCustomerId();
            $email      = $order->getCustomerEmail();
            $customer   = $this->_customerFactory->create()->load($customerId);
            $id         = $this->checkSalesforceCustomerStatus($customer);
            if ($id) {
                $accountIds[$orderId] = $id;
            } elseif ($customerId) {
                if (!isset($unsyncedOrderCustomerIds[$customerId])) {
                    $unsyncedOrderCustomerIds[$customerId] = [
                        'mid'  => $customerId,
                        'mObj' => $customer
                    ];
                }
                $unsyncedOrderCustomerIds[$customerId]['order_ids'][] = $orderId;
            } else {
                if (!isset($unsyncedEmails[$email])) {
                    $unsyncedEmails[$email] = [
                        'email'     => $email,
                        'firstName' => $order->getCustomerFirstname(),
                        'lastName'  => $order->getCustomerLastname()
                    ];
                }
                $unsyncedEmails[$email]['order_ids'][] = $orderId;

                $address = $order->getBillingAddress() ?: $order->getShippingAddress();
                $data    = [
                    'Email'     => $email,
                    'FirstName' => $address->getFirstname(),
                    'LastName'  => $address->getLastname(),
                ];

                $unsyncedDataEmails[$email] = $data;
            }
        }
        if (!empty($unsyncedOrderCustomerIds)) {
            $unsyncedCustomerIds = array_values($unsyncedOrderCustomerIds);
            $response            = $this->_account->sendAccountsRequest($unsyncedCustomerIds, 'insert');
            $this->_contact->sendContactsRequest($unsyncedCustomerIds, 'insert');
            if (is_array($response)) {
                $total = count($response);
                for ($i = 0; $i < $total; $i++) {
                    if (isset($response[$i]['success']) && $response[$i]['success']) {
                        foreach ($unsyncedOrderCustomerIds[$unsyncedCustomerIds[$i]['mid']]['order_ids'] as $order_id) {
                            $accountIds[$order_id] = $response[$i]['id'];
                        }
                    }
                }
            }
        }
        if (!empty($unsyncedEmails)) {
            $response = $this->_account->syncByEmails($unsyncedEmails, 'insert');
            $this->_contact->syncByEmails($unsyncedDataEmails, 'insert');
            $accountIds += $response;
        }

        return $accountIds;
    }

    /**
     * @return array|mixed|string
     * @throws \Exception
     */
    protected function createOrderItems()
    {
        if (empty($this->createOrderIds) && empty($this->updateOrderIds)) {
            return [];
        }
        $params   = [];
        $itemIds  = [];
        $orderIds = ($this->createOrderIds ?? []) + ($this->updateOrderIds ?? []);
        // Get pricebookId of "Standard Price Book"
        $pricebookId = $this->_getStandardPriceBookId();

        foreach ($orderIds as $orderId) {
            /** @var \Magento\Sales\Model\Order $order */
            $order   = $orderId['mObj'];
            $orderId = $this->checkExistedOrder($order);
            if (!empty($orderId)) {
                /** @var Item $item */
                foreach ($order->getAllVisibleItems() as $item) {
                    $product         = $item->getProduct();
                    $productEntityId = $item->getProductId();
                    $qty             = $item->getQtyOrdered();
                    $price           = $item->getPrice() / $qty;

                    if ($price > 0) {
                        $productId = $this->_product->syncOrderItem($item);
                        $pricebookEntryId = $this->getPricebookEntryId(
                            $productId,
                            $pricebookId,
                            $product ? ($product->getPrice() ?: 0) : ($item->getOriginalPrice() ?: 0),
                            $item->getProductId()
                        );
                        if ($productId && $pricebookEntryId) {
                            if (!$this->searchOrderItems('OrderItem', 'OrderId', $orderId, $pricebookEntryId, (double)$qty, (double)$price)) {
                                $params[]  = [
                                    'PricebookEntryId' => $pricebookEntryId,
                                    'OrderId'          => $orderId,
                                    'Quantity'         => $qty,
                                    'UnitPrice'        => $price,
                                ];
                                $itemIds[] = ['mid' => $productEntityId];
                            }
                        }
                    }
                }

                if ($orderId && in_array($orderId, $this->createOrderIds ?? [])) {
                    if ($taxInfo = $this->getTaxItemInfo($order, $orderId)) {
                        $params[]  = $taxInfo;
                        $itemIds[] = ['mid' => 'TAX'];
                    }
                    if ($shippingInfo = $this->getShippingItemInfo($order, $orderId)) {
                        $params[]  = $shippingInfo;
                        $itemIds[] = ['mid' => 'SHIPPING'];
                    }
                    if ($discountInfo = $this->getDiscountItemInfo($order, $orderId)) {
                        $params[]  = $discountInfo;
                        $itemIds[] = ['mid' => 'DISCOUNT'];
                    }
                }
            }
        }
        $response = $this->_job->sendBatchRequest('insert', 'OrderItem', $this->_json->serialize($params));
        $this->saveReports('create', 'OrderItem', $response, $itemIds, $params);
        return $response;
    }

    /**
     * @param \Magento\Sales\Model\Order $order
     * @param string $orderId
     * @return array|null
     */
    public function getTaxItemInfo($order, $orderId)
    {
        $taxAmount = $order->getTaxAmount();
        if ($taxAmount > 0) {
            $info = [
                'PricebookEntryId' => $this->_scopeConfig->getValue(Product::XML_TAX_PRICEBOOKENTRY_ID_PATH),
                'OrderId'          => $orderId,
                'Quantity'         => 1,
                'UnitPrice'        => $taxAmount,
            ];
            return $info;
        }
        return null;
    }

    /**
     * @param \Magento\Sales\Model\Order $order
     * @param string $orderId
     * @return array|null
     */
    public function getShippingItemInfo($order, $orderId)
    {
        $shippingAmount = $order->getShippingAmount();
        if ($shippingAmount > 0) {
            $info = [
                'PricebookEntryId' => $this->_scopeConfig->getValue(Product::XML_SHIPPING_PRICEBOOKENTRY_ID_PATH),
                'OrderId'          => $orderId,
                'Quantity'         => 1,
                'UnitPrice'        => $shippingAmount,
            ];
            return $info;
        }
        return null;
    }

    /**
     * @param \Magento\Sales\Model\Order $order
     * @param string $orderId
     * @return array|null
     */
    public function getDiscountItemInfo($order, $orderId)
    {
        $discountAmount = $order->getDiscountAmount();
        if ($discountAmount < 0) {
            $info = [
                'PricebookEntryId' => $this->_scopeConfig->getValue(Product::XML_DISCOUNT_PRICEBOOKENTRY_ID_PATH),
                'OrderId'          => $orderId,
                'Quantity'         => 1,
                'UnitPrice'        => $discountAmount,
            ];
            return $info;
        }
        return null;
    }

    /**
     * @param $orderIds
     * @param $response
     * @throws \Exception
     */
    protected function saveAttributes($orderIds, $response)
    {
        if (empty($orderIds)) {
            return;
        }

        if (is_array($response) && is_array($orderIds)) {
            $total = count($response);
            for ($i = 0; $i < $total; $i++) {
                if (isset($orderIds[$i])) {
                    $order = $orderIds[$i]['mObj'];
                    if (isset($response[$i]['id']) && $order->getId()) {
                        $this->saveAttribute($order, $response[$i]['id']);
                    }
                }
            }
        }
    }

    /**
     * @param \Magento\Sales\Model\Order $order
     * @param $salesforceId
     * @throws \Exception
     */
    protected function saveAttribute($order, $salesforceId)
    {
        try {
            $resource = $order->getResource();
            $order->setData(self::SALESFORCE_ORDER_ATTRIBUTE_CODE, $salesforceId);
            $resource->saveAttribute($order, self::SALESFORCE_ORDER_ATTRIBUTE_CODE);
        } catch (\Throwable $e) {
            $this->_logger->critical($e->getMessage(), ['trace' => $e->getTraceAsString()]);
        }
    }

    /**
     * @param \Magento\Sales\Model\Order $order
     * @return false|mixed
     * @throws \Zend_Http_Client_Exception
     */
    public function checkExistedOrder(\Magento\Sales\Model\Order $order)
    {
        $salesforceId     = false;
        $orderIncrementId = $order->getIncrementId();
        $orderData        = $order->getData();
        if (!empty($orderData) && isset($orderData[Order::SALESFORCE_ORDER_ATTRIBUTE_CODE])) {
            $salesforceId = $orderData[Order::SALESFORCE_ORDER_ATTRIBUTE_CODE];
        }
        if (!$salesforceId) {
            $orderInReport = $this->checkLogReportByMagentoId($orderIncrementId);
            if ($orderInReport->getData('record_id')) {
                return $orderInReport->getRecordId();
            }
            return $this->searchRecords($this->_type, 'Name', $orderIncrementId);
        } else {
            return $salesforceId;
        }
    }
}
